@@ -1,7 +1,7 @@
 /* LICENSE BEGIN
     This file is part of the SixtyFPS Project -- https://sixtyfps.io
-    Copyright (c) 2020 Olivier Goffart <olivier.goffart@sixtyfps.io>
-    Copyright (c) 2020 Simon Hausmann <simon.hausmann@sixtyfps.io>
+    Copyright (c) 2021 Olivier Goffart <olivier.goffart@sixtyfps.io>
+    Copyright (c) 2021 Simon Hausmann <simon.hausmann@sixtyfps.io>
 
     SPDX-License-Identifier: GPL-3.0-only
     This file is also available under commercial licensing terms.
@@ -9,6 +9,8 @@
 LICENSE END */
 //! This module contains the GraphicsWindow that used to be within corelib.
 
+// cspell:ignore corelib winit Borderless
+
 use core::cell::{Cell, RefCell};
 use core::pin::Pin;
 use std::rc::{Rc, Weak};
@@ -16,22 +18,25 @@ use std::rc::{Rc, Weak};
 use const_field_offset::FieldOffsets;
 use corelib::component::ComponentRc;
 use corelib::graphics::*;
-use corelib::input::{KeyboardModifiers, MouseEvent, MouseEventType};
+use corelib::input::{KeyboardModifiers, MouseEvent};
 use corelib::items::ItemRef;
+use corelib::layout::Orientation;
 use corelib::slice::Slice;
-use corelib::window::{ComponentWindow, PlatformWindow};
+use corelib::window::PlatformWindow;
 use corelib::Property;
 use corelib::SharedString;
 use sixtyfps_corelib as corelib;
+use winit::dpi::LogicalSize;
 
 /// FIXME! this is some remains from a time where the GLRenderer was called the backend
 type Backend = super::GLRenderer;
+type BackendItemGraphicsCache = super::ItemGraphicsCache;
+type BackendImageCache = super::ImageCache;
 
-type WindowFactoryFn =
-    dyn Fn(&dyn crate::eventloop::EventLoopInterface, winit::window::WindowBuilder) -> Backend;
+type WindowFactoryFn = dyn Fn(winit::window::WindowBuilder) -> Backend;
 
 /// GraphicsWindow is an implementation of the [PlatformWindow][`crate::eventloop::PlatformWindow`] trait. This is
-/// typically instantiated by entry factory functions of the different graphics backends.
+/// typically instantiated by entry factory functions of the different graphics back ends.
 pub struct GraphicsWindow {
     pub(crate) self_weak: Weak<corelib::window::Window>,
     window_factory: Box<WindowFactoryFn>,
@@ -45,6 +50,10 @@ pub struct GraphicsWindow {
     active_popup: std::cell::RefCell<Option<(ComponentRc, Point)>>,
 
     default_font_properties: Pin<Rc<Property<FontRequest>>>,
+
+    pub(crate) graphics_cache: RefCell<BackendItemGraphicsCache>,
+    // This cache only contains textures. The cache for decoded CPU side images is in crate::IMAGE_CACHE.
+    pub(crate) texture_cache: RefCell<BackendImageCache>,
 }
 
 impl GraphicsWindow {
@@ -56,8 +65,7 @@ impl GraphicsWindow {
     ///   backing window.
     pub(crate) fn new(
         window_weak: &Weak<corelib::window::Window>,
-        graphics_backend_factory: impl Fn(&dyn crate::eventloop::EventLoopInterface, winit::window::WindowBuilder) -> Backend
-            + 'static,
+        graphics_backend_factory: impl Fn(winit::window::WindowBuilder) -> Backend + 'static,
     ) -> Rc<Self> {
         let default_font_properties_prop = Rc::pin(Property::default());
         default_font_properties_prop.set_binding({
@@ -71,7 +79,7 @@ impl GraphicsWindow {
                         let component = ComponentRc::borrow_pin(&component_rc);
                         let root_item = component.as_ref().get_item_ref(0);
                         ItemRef::downcast_pin(root_item).map(
-                            |window_item: Pin<&corelib::items::Window>| {
+                            |window_item: Pin<&corelib::items::WindowItem>| {
                                 window_item.default_font_properties()
                             },
                         )
@@ -89,18 +97,24 @@ impl GraphicsWindow {
             mouse_input_state: Default::default(),
             active_popup: Default::default(),
             default_font_properties: default_font_properties_prop,
+            graphics_cache: Default::default(),
+            texture_cache: Default::default(),
         })
     }
 
-    fn apply_geometry_constraint(&self, constraints: corelib::layout::LayoutInfo) {
+    fn apply_geometry_constraint(
+        &self,
+        constraints_horizontal: corelib::layout::LayoutInfo,
+        constraints_vertical: corelib::layout::LayoutInfo,
+    ) {
         match &*self.map_state.borrow() {
             GraphicsWindowBackendState::Unmapped => {}
             GraphicsWindowBackendState::Mapped(window) => {
-                if constraints != window.constraints.get() {
-                    let min_width = constraints.min_width.min(constraints.max_width);
-                    let min_height = constraints.min_height.min(constraints.max_height);
-                    let max_width = constraints.max_width.max(constraints.min_width);
-                    let max_height = constraints.max_height.max(constraints.min_height);
+                if (constraints_horizontal, constraints_vertical) != window.constraints.get() {
+                    let min_width = constraints_horizontal.min.min(constraints_horizontal.max);
+                    let min_height = constraints_vertical.min.min(constraints_vertical.max);
+                    let max_width = constraints_horizontal.max.max(constraints_horizontal.min);
+                    let max_height = constraints_vertical.max.max(constraints_vertical.min);
 
                     window.backend.borrow().window().set_min_inner_size(
                         if min_width > 0. || min_height > 0. {
@@ -119,7 +133,7 @@ impl GraphicsWindow {
                             None
                         },
                     );
-                    window.constraints.set(constraints);
+                    window.constraints.set((constraints_horizontal, constraints_vertical));
 
                     #[cfg(target_arch = "wasm32")]
                     {
@@ -146,10 +160,19 @@ impl GraphicsWindow {
         }
     }
 
+    fn with_current_context<T>(&self, cb: impl FnOnce() -> T) -> T {
+        match &*self.map_state.borrow() {
+            GraphicsWindowBackendState::Unmapped => cb(),
+            GraphicsWindowBackendState::Mapped(window) => {
+                window.backend.borrow().with_current_context(cb)
+            }
+        }
+    }
+
     /// Requests for the window to be mapped to the screen.
     ///
     /// Arguments:
-    /// * `component`: The component that holds the root item of the scene. If the item is a [`corelib::items::Window`], then
+    /// * `component`: The component that holds the root item of the scene. If the item is a [`corelib::items::WindowItem`], then
     ///   the `width` and `height` properties are read and the values are passed to the windowing system as request
     ///   for the initial size of the window. Then bindings are installed on these properties to keep them up-to-date
     ///   with the size as it may be changed by the user or the windowing system in general.
@@ -162,51 +185,59 @@ impl GraphicsWindow {
         let component = ComponentRc::borrow_pin(&component);
         let root_item = component.as_ref().get_item_ref(0);
 
-        let window_title =
-            if let Some(window_item) = ItemRef::downcast_pin::<corelib::items::Window>(root_item) {
-                window_item.title().to_string()
-            } else {
-                "SixtyFPS Window".to_string()
-            };
+        let window_title = if let Some(window_item) =
+            ItemRef::downcast_pin::<corelib::items::WindowItem>(root_item)
+        {
+            window_item.title().to_string()
+        } else {
+            "SixtyFPS Window".to_string()
+        };
         let window_builder = winit::window::WindowBuilder::new().with_title(window_title);
 
-        let id = {
-            let backend = crate::eventloop::with_window_target(|event_loop| {
-                self.window_factory.as_ref()(event_loop, window_builder)
-            });
-
-            // Ideally we should be passing the initial requested size to the window builder, but those properties
-            // may be specified in logical pixels, relative to the scale factory, which we only know *after* mapping
-            // the window to the screen. So we first map the window then, propagate the scale factory and *then* the
-            // width/height properties should have the correct values calculated via their bindings that multiply with
-            // the scale factor.
-            // We could pass the logical requested size at window builder time, *if* we knew what the values are.
-            let window_id = {
-                let platform_window = backend.window();
-
-                if std::env::var("SIXTYFPS_FULLSCREEN").is_ok() {
-                    platform_window
-                        .set_fullscreen(Some(winit::window::Fullscreen::Borderless(None)));
-                }
-
-                self.properties.as_ref().scale_factor.set(platform_window.scale_factor() as _);
-
-                platform_window.id()
-            };
-
-            self.map_state.replace(GraphicsWindowBackendState::Mapped(MappedWindow {
-                backend: RefCell::new(backend),
-                constraints: Default::default(),
-            }));
-
-            window_id
+        let window_builder = if std::env::var("SIXTYFPS_FULLSCREEN").is_ok() {
+            window_builder.with_fullscreen(Some(winit::window::Fullscreen::Borderless(None)))
+        } else {
+            let component_rc = self.component();
+            let component = ComponentRc::borrow_pin(&component_rc);
+            let layout_info_h = component.as_ref().layout_info(Orientation::Horizontal);
+            let layout_info_v = component.as_ref().layout_info(Orientation::Vertical);
+            let s = LogicalSize::new(
+                layout_info_h.preferred_bounded(),
+                layout_info_v.preferred_bounded(),
+            );
+            if s.width > 0. && s.height > 0. {
+                // Make sure that the window's inner size is in sync with the root window item's
+                // width/height.
+                self.set_geometry(s.width, s.height);
+                window_builder.with_inner_size(s)
+            } else {
+                window_builder
+            }
         };
 
-        crate::eventloop::register_window(id, self.clone());
+        let backend = self.window_factory.as_ref()(window_builder);
+
+        let platform_window = backend.window();
+        self.properties.as_ref().scale_factor.set(platform_window.scale_factor() as _);
+        let id = platform_window.id();
+        drop(platform_window);
+
+        self.map_state.replace(GraphicsWindowBackendState::Mapped(MappedWindow {
+            backend: RefCell::new(backend),
+            constraints: Default::default(),
+        }));
+
+        crate::event_loop::register_window(id, self);
     }
     /// Removes the window from the screen. The window is not destroyed though, it can be show (mapped) again later
     /// by calling [`PlatformWindow::map_window`].
     fn unmap_window(self: Rc<Self>) {
+        // Release GL textures and other GPU bound resources.
+        self.with_current_context(|| {
+            self.graphics_cache.borrow_mut().clear();
+            self.texture_cache.borrow_mut().remove_textures();
+        });
+
         self.map_state.replace(GraphicsWindowBackendState::Unmapped);
         /* FIXME:
         if let Some(existing_blinker) = self.cursor_blinker.borrow().upgrade() {
@@ -232,35 +263,35 @@ impl GraphicsWindow {
             let component = ComponentRc::borrow_pin(&component_rc);
 
             runtime_window.meta_properties_tracker.as_ref().evaluate_if_dirty(|| {
-                self.apply_geometry_constraint(component.as_ref().layout_info());
-                component.as_ref().apply_layout(Default::default());
+                self.apply_geometry_constraint(
+                    component.as_ref().layout_info(Orientation::Horizontal),
+                    component.as_ref().layout_info(Orientation::Vertical),
+                );
 
-                if let Some((popup, pos)) = &*self.active_popup.borrow() {
+                if let Some((popup, _)) = &*self.active_popup.borrow() {
                     let popup = ComponentRc::borrow_pin(popup);
                     let popup_root = popup.as_ref().get_item_ref(0);
-                    let size = if let Some(window_item) = ItemRef::downcast_pin(popup_root) {
-                        let layout_info = popup.as_ref().layout_info();
+                    if let Some(window_item) = ItemRef::downcast_pin(popup_root) {
+                        let layout_info_h = popup.as_ref().layout_info(Orientation::Horizontal);
+                        let layout_info_v = popup.as_ref().layout_info(Orientation::Vertical);
 
                         let width =
-                            corelib::items::Window::FIELD_OFFSETS.width.apply_pin(window_item);
+                            corelib::items::WindowItem::FIELD_OFFSETS.width.apply_pin(window_item);
                         let mut w = width.get();
-                        if w < layout_info.min_width {
-                            w = layout_info.min_width;
-                            width.set(w);
-                        }
-
                         let height =
-                            corelib::items::Window::FIELD_OFFSETS.height.apply_pin(window_item);
+                            corelib::items::WindowItem::FIELD_OFFSETS.height.apply_pin(window_item);
                         let mut h = height.get();
-                        if h < layout_info.min_height {
-                            h = layout_info.min_height;
-                            height.set(h);
+                        if w <= 0. {
+                            w = layout_info_h.preferred;
                         }
-                        Size::new(h, w) * self.scale_factor()
-                    } else {
-                        Size::default()
+                        if h <= 0. {
+                            h = layout_info_v.preferred;
+                        }
+                        w = w.clamp(layout_info_h.min, layout_info_h.max);
+                        h = h.clamp(layout_info_v.min, layout_info_v.max);
+                        width.set(w);
+                        height.set(h);
                     };
-                    popup.as_ref().apply_layout(Rect::new(pos.clone(), size));
                 }
             });
 
@@ -268,14 +299,15 @@ impl GraphicsWindow {
             let window = map_state.as_mapped();
             let root_item = component.as_ref().get_item_ref(0);
             let background_color = if let Some(window_item) =
-                ItemRef::downcast_pin::<corelib::items::Window>(root_item)
+                ItemRef::downcast_pin::<corelib::items::WindowItem>(root_item)
             {
                 window_item.background()
             } else {
-                RgbaColor { red: 255 as u8, green: 255, blue: 255, alpha: 255 }.into()
+                RgbaColor { red: 255_u8, green: 255, blue: 255, alpha: 255 }.into()
             };
 
             let mut renderer = window.backend.borrow_mut().new_renderer(
+                self.clone(),
                 &background_color,
                 self.scale_factor(),
                 self.default_font_properties(),
@@ -293,17 +325,17 @@ impl GraphicsWindow {
     }
 
     /// FIXME: this is the same as Window::process_mouse_input, but this handle the popup.
-    /// Ideally the popup should be handled as a different window or by theevent loop, and
+    /// Ideally the popup should be handled as a different window or by the event loop, and
     /// this function can go away
-    pub fn process_mouse_input(self: Rc<Self>, mut pos: Point, what: MouseEventType) {
+    pub fn process_mouse_input(self: Rc<Self>, mut event: MouseEvent) {
         let active_popup = (*self.active_popup.borrow()).clone();
         let component = if let Some(popup) = &active_popup {
-            pos -= popup.1.to_vector();
-            if what == MouseEventType::MousePressed {
+            event.translate(-popup.1.to_vector());
+            if let MouseEvent::MousePressed { pos } = &event {
                 // close the popup if one press outside the popup
                 let geom =
                     ComponentRc::borrow_pin(&popup.0).as_ref().get_item_ref(0).as_ref().geometry();
-                if !geom.contains(pos) {
+                if !geom.contains(*pos) {
                     self.close_popup();
                     return;
                 }
@@ -315,15 +347,15 @@ impl GraphicsWindow {
 
         self.mouse_input_state.set(corelib::input::process_mouse_input(
             component,
-            MouseEvent { pos, what },
-            &ComponentWindow::new(self.self_weak.upgrade().unwrap()),
+            event,
+            &self.self_weak.upgrade().unwrap(),
             self.mouse_input_state.take(),
         ));
 
         if active_popup.is_some() {
             //FIXME: currently the ComboBox is the only thing that uses the popup, and it should close automatically
             // on release.  But ideally, there would be API to close the popup rather than always closing it on release
-            if what == MouseEventType::MouseReleased {
+            if matches!(event, MouseEvent::MouseReleased { .. }) {
                 self.close_popup();
             }
         }
@@ -354,14 +386,16 @@ impl GraphicsWindow {
     /// window resize event from the windowing system.
     /// Size is in logical pixels.
     pub fn set_geometry(&self, width: f32, height: f32) {
-        self.self_weak.upgrade().unwrap().try_component().map(|component_rc| {
+        if let Some(component_rc) = self.self_weak.upgrade().unwrap().try_component() {
             let component = ComponentRc::borrow_pin(&component_rc);
             let root_item = component.as_ref().get_item_ref(0);
-            if let Some(window_item) = ItemRef::downcast_pin::<corelib::items::Window>(root_item) {
+            if let Some(window_item) =
+                ItemRef::downcast_pin::<corelib::items::WindowItem>(root_item)
+            {
                 window_item.width.set(width);
                 window_item.height.set(height);
             }
-        });
+        }
     }
 }
 
@@ -383,15 +417,21 @@ impl PlatformWindow for GraphicsWindow {
         self.properties.as_ref().scale_factor.set(factor);
     }
 
-    fn free_graphics_resources<'a>(self: Rc<Self>, items: &Slice<'a, Pin<ItemRef<'a>>>) {
+    fn free_graphics_resources<'a>(&self, items: &Slice<'a, Pin<ItemRef<'a>>>) {
         match &*self.map_state.borrow() {
             GraphicsWindowBackendState::Unmapped => {}
-            GraphicsWindowBackendState::Mapped(window) => {
-                for item in items.iter() {
-                    let cached_rendering_data = item.cached_rendering_data_offset();
-                    cached_rendering_data.release(
-                        &mut window.backend.borrow().shared_data.item_graphics_cache.borrow_mut(),
-                    )
+            GraphicsWindowBackendState::Mapped(_) => {
+                let mut cache_entries_to_clear = items
+                    .iter()
+                    .flat_map(|item| {
+                        let cached_rendering_data = item.cached_rendering_data_offset();
+                        self.graphics_cache.borrow_mut().release(cached_rendering_data)
+                    })
+                    .peekable();
+                if cache_entries_to_clear.peek().is_some() {
+                    self.with_current_context(|| {
+                        cache_entries_to_clear.for_each(drop);
+                    });
                 }
             }
         }
@@ -399,7 +439,9 @@ impl PlatformWindow for GraphicsWindow {
 
     fn show_popup(&self, popup: &ComponentRc, position: Point) {
         *self.active_popup.borrow_mut() = Some((popup.clone(), position));
-        self.self_weak.upgrade().map(|window| window.meta_properties_tracker.set_dirty());
+        if let Some(window) = self.self_weak.upgrade() {
+            window.meta_properties_tracker.set_dirty();
+        }
     }
 
     fn close_popup(&self) {
@@ -411,14 +453,14 @@ impl PlatformWindow for GraphicsWindow {
         match &*self.map_state.borrow() {
             GraphicsWindowBackendState::Unmapped => {
                 // Nothing to be done if the window isn't visible. When it becomes visible,
-                // ComponentWindow::show() calls update_window_properties()
+                // corelib::window::Window::show() calls update_window_properties()
             }
             GraphicsWindowBackendState::Mapped(window) => {
                 let backend = window.backend.borrow();
                 let window_id = backend.window().id();
-                crate::eventloop::with_window_target(|event_loop| {
+                crate::event_loop::with_window_target(|event_loop| {
                     event_loop.event_loop_proxy().send_event(
-                        crate::eventloop::CustomEvent::UpdateWindowProperties(window_id),
+                        crate::event_loop::CustomEvent::UpdateWindowProperties(window_id),
                     )
                 })
                 .ok();
@@ -426,25 +468,55 @@ impl PlatformWindow for GraphicsWindow {
         }
     }
 
-    fn apply_window_properties(&self, window_item: Pin<&sixtyfps_corelib::items::Window>) {
+    fn apply_window_properties(&self, window_item: Pin<&sixtyfps_corelib::items::WindowItem>) {
         match &*self.map_state.borrow() {
             GraphicsWindowBackendState::Unmapped => {}
             GraphicsWindowBackendState::Mapped(window) => {
-                let backend = window.backend.borrow();
-                let winit_window = backend.window();
-                winit_window.set_title(&window_item.title());
-                let mut size = winit_window.inner_size().to_logical(self.scale_factor() as f64);
+                let title = window_item.title();
+                let icon = window_item.icon();
+                let mut size: LogicalSize<f64> = {
+                    let backend = window.backend.borrow();
+                    let winit_window = backend.window();
+                    winit_window.set_title(&title);
+                    if let Some(rgba) = crate::IMAGE_CACHE
+                        .with(|c| c.borrow_mut().load_image_resource((&icon).into()))
+                        .and_then(|i| i.to_rgba())
+                    {
+                        let (width, height) = rgba.dimensions();
+                        winit_window.set_window_icon(
+                            winit::window::Icon::from_rgba(rgba.into_raw(), width, height).ok(),
+                        );
+                    };
+                    winit_window.inner_size().to_logical(self.scale_factor() as f64)
+                };
                 let mut must_resize = false;
-                let mut apply = |r: &mut u32, v| {
-                    if v > 0. {
-                        *r = v as _
-                    } else {
-                        must_resize = true
+                let mut w = window_item.width();
+                let mut h = window_item.height();
+                if (size.width as f32 - w).abs() < 1. || (size.height as f32 - h).abs() < 1. {
+                    return;
+                }
+                if w <= 0. || h <= 0. {
+                    if let Some(component_rc) = self.self_weak.upgrade().unwrap().try_component() {
+                        let component = ComponentRc::borrow_pin(&component_rc);
+                        if w <= 0. {
+                            let info = component.as_ref().layout_info(Orientation::Horizontal);
+                            w = info.preferred_bounded();
+                            must_resize = true;
+                        }
+                        if h <= 0. {
+                            let info = component.as_ref().layout_info(Orientation::Vertical);
+                            h = info.preferred_bounded();
+                            must_resize = true;
+                        }
                     }
                 };
-                apply(&mut size.width, window_item.width());
-                apply(&mut size.height, window_item.height());
-                winit_window.set_inner_size(size);
+                if w > 0. {
+                    size.width = w as _;
+                }
+                if h > 0. {
+                    size.height = h as _;
+                }
+                window.backend.borrow().window().set_inner_size(size);
                 if must_resize {
                     self.set_geometry(size.width as _, size.height as _)
                 }
@@ -462,38 +534,24 @@ impl PlatformWindow for GraphicsWindow {
 
     fn font_metrics(
         &self,
-        item_graphics_cache: &corelib::item_rendering::CachedRenderingData,
+        item_graphics_cache_data: &corelib::item_rendering::CachedRenderingData,
         unresolved_font_request_getter: &dyn Fn() -> corelib::graphics::FontRequest,
         reference_text: Pin<&Property<SharedString>>,
-    ) -> Option<Box<dyn corelib::graphics::FontMetrics>> {
-        match &*self.map_state.borrow() {
-            GraphicsWindowBackendState::Unmapped => None,
-            GraphicsWindowBackendState::Mapped(window) => Some(
-                window.backend.borrow_mut().font_metrics(
-                    item_graphics_cache,
-                    &|| {
-                        unresolved_font_request_getter()
-                            .merge(&self.default_font_properties().as_ref().get())
-                    },
-                    WindowProperties::FIELD_OFFSETS
-                        .scale_factor
-                        .apply_pin(self.properties.as_ref()),
-                    reference_text,
-                ),
-            ),
-        }
-    }
+    ) -> Box<dyn corelib::graphics::FontMetrics> {
+        let font_request_fn = || {
+            unresolved_font_request_getter().merge(&self.default_font_properties().as_ref().get())
+        };
 
-    fn image_size(
-        &self,
-        source: Pin<&sixtyfps_corelib::properties::Property<ImageReference>>,
-    ) -> sixtyfps_corelib::graphics::Size {
-        match &*self.map_state.borrow() {
-            GraphicsWindowBackendState::Unmapped => Default::default(),
-            GraphicsWindowBackendState::Mapped(window) => {
-                window.backend.borrow().image_size(source)
-            }
-        }
+        let scale_factor =
+            WindowProperties::FIELD_OFFSETS.scale_factor.apply_pin(self.properties.as_ref());
+
+        Box::new(super::fonts::FontMetrics::new(
+            &mut *self.graphics_cache.borrow_mut(),
+            item_graphics_cache_data,
+            font_request_fn,
+            scale_factor,
+            reference_text,
+        ))
     }
 
     fn as_any(&self) -> &dyn std::any::Any {
@@ -503,12 +561,12 @@ impl PlatformWindow for GraphicsWindow {
 
 struct MappedWindow {
     backend: RefCell<Backend>,
-    constraints: Cell<corelib::layout::LayoutInfo>,
+    constraints: Cell<(corelib::layout::LayoutInfo, corelib::layout::LayoutInfo)>,
 }
 
 impl Drop for MappedWindow {
     fn drop(&mut self) {
-        crate::eventloop::unregister_window(self.backend.borrow().window().id());
+        crate::event_loop::unregister_window(self.backend.borrow().window().id());
     }
 }
 
@@ -523,7 +581,7 @@ impl GraphicsWindowBackendState {
             GraphicsWindowBackendState::Unmapped => panic!(
                 "internal error: tried to access window functions that require a mapped window"
             ),
-            GraphicsWindowBackendState::Mapped(mw) => &mw,
+            GraphicsWindowBackendState::Mapped(mw) => mw,
         }
     }
 }
